/* SPDX-License-Identifier: BSD-2-Clause */

/*
 * This file defines the API for the sip module's methods.
 *
 * Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>
 */


#ifndef _SIP_SIP_MODULE_METHODS_H
#define _SIP_SIP_MODULE_METHODS_H

#include <Python.h>


#ifdef __cplusplus
extern "C" {
#endif

extern PyMethodDef sipSipModuleMethods[];

#ifdef __cplusplus
}
#endif

#endif
