/* SPDX-License-Identifier: BSD-2-Clause */

/*
 * Forward declarations of internal data structures.
 *
 * Copyright (c) 2026 Phil Thompson <phil@riverbankcomputing.com>
 */


#ifndef _SIP_DECLS_H
#define _SIP_DECLS_H

struct sipSipModuleStateImpl;
typedef struct sipSipModuleStateImpl sipSipModuleState;

#endif
